/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */


/*
 * MathLib.java
 *
 * Created on 07 September 2008, 10:34
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package net.infantai.core;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;



/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class MathLib
{
    
    /**
     * Creates a new instance of MathLibrary.
     */
    public MathLib() 
    {
        // All methods will be static, nothing needed in the default constructor.
    }
    
    /**
     * Computes the Quadratic Mean (Root Mean Square) of the input values.
     */
    public static double computeQuadraticMean(double[] values)
    {
        double rms = 0.0;
        for (int i = 0; i < values.length; ++i)
        {
            rms += square(values[i]);
        }
        rms = rms / (double) values.length;
        rms = Math.sqrt(rms);
        return rms;
    }
    
    /**
     * Returns the square of the input value.
     */
    public static double square(double value)
    {
        return (value * value);
    }
    
    /**
     * Computes the Variance of the input values.
     */
    public static double computeVariance(double[] values)
    {
        double variance = 0.0;
        double mean = computeMean(values);
        for (int i = 0; i < values.length; ++i)
        {
            variance = variance + square((values[i] - mean));
        }
        variance = variance / values.length;
        return variance;
    }
    
    /**
     * Computes the Standard Deviation of the input values. Standard Deviation is the square root of the Variance.
     */
    public static double computeStandardDeviation(double[] values)
    {
        return Math.sqrt(computeVariance(values));
    }
    
    /**
     * Computes the Mean (Average) of the input values.
     */
    public static double computeMean(double[] values)
    {
        double mean = 0.0;
        for (int i = 0; i < values.length; ++i)
        {
            mean = mean + values[i];
        }
        mean = mean / values.length;
        return mean;
    }
    
    /**
     * Performs a scientific rounding of the input value.
     */
    public static double round(double value, int decimalPrecision)
    {
        double multiplier = Math.pow(10, decimalPrecision + 1.0);
        if (value >= (Double.MAX_VALUE / multiplier))
        {
            return value;
        }
        int i1 = 0;
        int i2 = 0;
        double d1 = 0.0;
        double d2 = 0.0;
        d1 = value * multiplier;
        i1 = (int) (d1 / 10);
        i2 = (int) (d1 - (i1 * 10));
        if (i2 >= 5)
        {
            i1++;
        }
        d2 = (double) (i1 * 10.0) / multiplier;
        return d2;
    }
    
    /**
     * Computes and returns the Euclidean Norm of the length of the input vector.
     */
    public static double computeTFVectorLengthNorm(Map<String, Double> fv)
    {
        double vectorLength = 0.0;
        Iterator<Map.Entry<String, Double>> iter = fv.entrySet().iterator();
        while (iter.hasNext())
        {
            Map.Entry<String, Double> mapEntry = iter.next();
            vectorLength += MathLib.square(fv.get(mapEntry.getKey()).doubleValue());
        }
        return Math.sqrt(vectorLength);
    }     
    
    /**
     * Normalizes the input Vector on the basis of its Euclidean Norm.
     */
    public static Map<String, Double> normalizeTermFrequencyVector(Map<String, Double> fv)
    {
        Map<String, Double> normalizedTFVector = new LinkedHashMap<String, Double>();
        double vectorLength = computeTFVectorLengthNorm(fv);
        Iterator<Map.Entry<String, Double>> iter = fv.entrySet().iterator();
        while (iter.hasNext())
        {
            Map.Entry<String, Double> mapEntry = iter.next();
            String term = mapEntry.getKey();
            Double freq = mapEntry.getValue();
            freq = freq / vectorLength;
            normalizedTFVector.put(term, freq);
        }
        return normalizedTFVector;
    }    
    
}
