/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

 /*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp;

import net.infantai.util.common.TokenGroup;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.GregorianCalendar;
import java.util.LinkedHashMap;
import java.util.Map;
import opennlp.tools.postag.POSModel;
import opennlp.tools.postag.POSTaggerME;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
// Penn Treebank Abbreviations: http://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html (University of Pennsylvania, Philadelphia, Pennsylvania, USA)
//Number  Tag     Description
//1.	CC	Coordinating conjunction
//2.	CD	Cardinal number
//3.	DT	Determiner
//4.	EX	Existential there
//5.	FW	Foreign word
//6.	IN	Preposition or subordinating conjunction
//7.	JJ	Adjective
//8.	JJR	Adjective, comparative
//9.	JJS	Adjective, superlative
//10.	LS	List item marker
//11.	MD	Modal
//12.	NN	Noun, singular or mass
//13.	NNS	Noun, plural
//14.	NNP	Proper noun, singular
//15.	NNPS	Proper noun, plural
//16.	PDT	Predeterminer
//17.	POS	Possessive ending
//18.	PRP	Personal pronoun
//19.	PRP$	Possessive pronoun
//20.	RB	Adverb
//21.	RBR	Adverb, comparative
//22.	RBS	Adverb, superlative
//23.	RP	Particle
//24.	SYM	Symbol
//25.	TO	to
//26.	UH	Interjection
//27.	VB	Verb, base form
//28.	VBD	Verb, past tense
//29.	VBG	Verb, gerund or present participle
//30.	VBN	Verb, past participle
//31.	VBP	Verb, non-3rd person singular present
//32.	VBZ	Verb, 3rd person singular present
//33.	WDT	Wh-determiner
//34.	WP	Wh-pronoun
//35.	WP$	Possessive wh-pronoun
//36.	WRB	Wh-adverb

public class POSTagger
{

    private static final Logger logger = LogManager.getLogger(POSTagger.class);

    private POSModel model = null;

    private long t1, t2;
    
    private static final String ERR_MSG_TAG_COUNT_MISMATCH = "Word count does not match tag count: ";

    public POSTagger()
    {
        initialize("./models/opennlp/en-pos-maxent.bin"); // from http://opennlp.sourceforge.net/models-1.5/
    }

    private void initialize(String lexiconFileName)
    {
        try
        {
            InputStream modelInStream = new FileInputStream(lexiconFileName);
            model = new POSModel(modelInStream);
        }
        catch (IOException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
            logger.error(ex.getMessage(), ex);
        }
    }

    public Map<String, String> tag(String text, int maxGroupCount, boolean allowPrepositions, boolean allowConjunctions, boolean allowAdjectives, boolean allowDeterminers, boolean allowAdverbs)
    {
        Map<String, String> groupedEntities = null;
        try
        {
            if (model != null)
            {
                POSTaggerME taggerME = new POSTaggerME(model);
                t1 = System.nanoTime();
                String[] words = text.split("\\W+");
                t2 = System.nanoTime();
                String[] tags = taggerME.tag(words);
                t2 = System.nanoTime();
                logger.trace("POS Tagging took " + ((double) (t2 - t1)) / 1000000.0 + " ms.");
                for (int i = 0; i < tags.length; i++)
                {
                    String word = words[i].trim();
                    String tag = tags[i].trim();
                    System.out.print(tag + ":" + word + "  ");
                }
//                LOG.debug("-----------------\n");
                groupedEntities = groupRelatedEntities(words, tags, maxGroupCount, allowPrepositions, allowConjunctions, allowAdjectives, allowDeterminers, allowAdverbs);
//                groupedEntities.forEach((key, value) -> 
//                        {
//                            LOG.debug(key + " => " + value);
//                });
            }
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
            logger.error(ex.getMessage(), ex);
        }
        return groupedEntities;
    }

    public Map<TokenGroup, String> tag(TokenGroup[] tgWords, int maxGroupLength, boolean allowPrepositions, boolean allowConjunctions, boolean allowAdjectives, boolean allowDeterminers, boolean allowAdverbs)
    {
        Map<TokenGroup, String> groupedEntities = null;
        try
        {
            if (model != null)
            {
                POSTaggerME taggerME = new POSTaggerME(model);
                t1 = System.nanoTime();
                String[] words = convertTokenGroupArrayToStringArray(tgWords);
                String[] tags = taggerME.tag(words);
                t2 = System.nanoTime();
                logger.trace("POS Tagging took " + ((double) (t2 - t1)) / 1000000.0 + " ms.");
//                for (int i = 0; i < tags.length; i++)
//                {
//                    String word = tgWords[i].getToken();
//                    String tag = tags[i].trim();
//                    LOG.debug(tag + ":" + word + "  ");
//                }
                groupedEntities = groupRelatedEntities(tgWords, tags, maxGroupLength, allowPrepositions, allowConjunctions, allowAdjectives, allowDeterminers, allowAdverbs);
//                groupedEntities.forEach((key, value) -> 
//                        {
//                            LOG.debug(key + " => " + value);
//                });
            }
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
            logger.error(ex.getMessage(), ex);
        }
        return groupedEntities;
    }

    public String[] convertTokenGroupArrayToStringArray(TokenGroup[] tgArr)
    {
        String[] words = new String[tgArr.length];
        for (int i = 0; i < words.length; ++i)
        {
            words[i] = tgArr[i].getToken();
        }
        return words;
    }
    
    public Map<String, String> groupRelatedEntities(String[] words, String[] tags, int maxGroupLength, boolean allowPrepositions, boolean allowConjunctions, boolean allowAdjectives, boolean allowDeterminers, boolean allowAdverbs)
    {
        Map<String, String> groupedEntities = new LinkedHashMap<String, String>();
        if (words.length != tags.length)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ERR_MSG_TAG_COUNT_MISMATCH + words.length + ", " + tags.length);
            logger.error(ERR_MSG_TAG_COUNT_MISMATCH + words.length + ", " + tags.length);
            return null;
        }
        for (int i = 0; i < words.length; ++i)
        {
            String tag = tags[i];
            int j = 1;
            if (tag.startsWith("NN") || tag.equals("CD")) // group all nouns, cardinal numbers
            {
                String reconstructedNoun = words[i];
                if (tag.startsWith("NN"))
                {
                    groupedEntities.put(reconstructedNoun, tag);
                }

                int groupCount = 1;

//                    while ((tags[i + j].startsWith("NN") || tags[i + j].equals("CD") || tags[i + j].equals("IN") || tags[i + j].equals("CC") || tags[i + j].startsWith("JJ")))
                while ((i + j) < words.length)
                {
                    if (maxGroupLength > -1 && groupCount >= maxGroupLength)
                    {
                        break;
                    }

                    boolean allowGroupingOfNextToken = false;

                    if (tags[i + j].startsWith("NN") || tags[i + j].equals("CD"))
                    {
                        allowGroupingOfNextToken = true; // always allow consecutive nouns to be grouped
                    }
                    else
                    {
                        if (allowPrepositions)
                        {
                            if (tags[i + j].equals("IN"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }

                        if (allowConjunctions)
                        {
                            if (tags[i + j].equals("CC")) // allow co-ordinating conjunctions
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }                            

                        if (allowAdjectives)
                        {
                            if (tags[i + j].startsWith("JJ"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }

                        if (allowDeterminers)
                        {
                            if (tags[i + j].equals("DT"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }

                        if (allowAdverbs)
                        {
                            if (tags[i + j].startsWith("RB"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }                            
                    }

                    if (!allowGroupingOfNextToken)
                    {
                        break;
                    }

                    reconstructedNoun = reconstructedNoun + ' ' + words[i + j];
                    j++;
                    groupCount++;
                    groupedEntities.put(reconstructedNoun, "Grouped NN");
                    if ((i + j) >= words.length)
                    {
                        break;
                    }
                }
                
            }
        }
        return groupedEntities;
    }

    public Map<TokenGroup, String> groupRelatedEntities(TokenGroup[] words, String[] tags, int maxGroupLength, boolean allowPrepositions, boolean allowConjunctions, boolean allowAdjectives, boolean allowDeterminers, boolean allowAdverbs)
    {
        Map<TokenGroup, String> groupedEntities = new LinkedHashMap<>();
        if (words.length != tags.length)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ERR_MSG_TAG_COUNT_MISMATCH + words.length + ", " + tags.length);
            logger.error(ERR_MSG_TAG_COUNT_MISMATCH + words.length + ", " + tags.length);
            return null;
        }
        for (int i = 0; i < words.length; ++i)
        {
            String tag = tags[i];
            int j = 1;
            if (tag.startsWith("NN") || tag.equals("CD")) // group all nouns, or cardinal numbers
            {
                if (tag.startsWith("NN"))
                {
                    TokenGroup reconstructedNoun = new TokenGroup(words[i].getToken(), words[i].getIndex(), words[i].getSize());
                    groupedEntities.put(reconstructedNoun, tag);
                }

                TokenGroup prevTG = null;
                int groupCount = 1;
                while ((i + j) < words.length)
                {
                    if (maxGroupLength > -1 && groupCount >= maxGroupLength)
                    {
                        break;
                    }

                    boolean allowGroupingOfNextToken = false;

                    if (tags[i + j].startsWith("NN") || tags[i + j].equals("CD"))
                    {
                        allowGroupingOfNextToken = true; // always allow consecutive nouns to be grouped
                    }
                    else
                    {
                        if (allowPrepositions)
                        {
                            if (tags[i + j].equals("IN"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }

                        if (allowConjunctions)
                        {
                            if (tags[i + j].equals("CC")) // allow co-ordinating conjunctions
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }                            

                        if (allowAdjectives)
                        {
                            if (tags[i + j].startsWith("JJ"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }

                        if (allowDeterminers)
                        {
                            if (tags[i + j].equals("DT"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }

                        if (allowAdverbs)
                        {
                            if (tags[i + j].startsWith("RB"))
                            {
                                allowGroupingOfNextToken = true;
                            }
                        }                            
                    }

                    if (!allowGroupingOfNextToken)
                    {
                        break;
                    }

                    TokenGroup reconstructedNoun = null;
                    if (prevTG != null)
                    {
                        reconstructedNoun = new TokenGroup(prevTG.getToken(), words[i].getIndex(), words[i].getSize());
                    }
                    else
                    {
                        reconstructedNoun = new TokenGroup(words[i].getToken(), words[i].getIndex(), words[i].getSize());
                    }
                    String originalToken = reconstructedNoun.getToken();
                    reconstructedNoun.setToken(originalToken + ' ' + words[i + j]);
                    reconstructedNoun.setSize(reconstructedNoun.getToken().length());
                    j++;
                    groupCount++;
                    groupedEntities.put(reconstructedNoun, "Grouped NN");
                    prevTG = reconstructedNoun;
                    if ((i + j) >= words.length)
                    {
                        break;
                    }
                }

            }
        }
        return groupedEntities;
    }
    
    public static void main(String[] args)
    {
        String input = "Pt is 87 yo woman, highschool teacher with past medical history that includes\n" +
"   - status post cardiac catheterization in April 2019.\n" +
"She presents today with palpitations and chest pressure.\n" +
"HPI : Sleeping trouble on present dosage of Clonidine. Severe Rash  on face and leg, slightly itchy  \n" +
"Meds : Vyvanse 50 mgs po at breakfast daily, \n" +
"            Clonidine 0.2 mgs -- 1 and 1 / 2 tabs po qhs \n" +
"HEENT : Boggy inferior turbinates, No oropharyngeal lesion \n" +
"Lungs : clear \n" +
"Heart : Regular rhythm \n" +
"Skin :  Mild erythematous eruption to hairline \n" +
"\n" +
"Follow-up as scheduled";
        POSTagger posTagger = new POSTagger();
        Map<String, String> results = posTagger.tag(input, 0, true, true, true, true, true);
        for (Map.Entry<String, String> mapEntry : results.entrySet())
        {
            System.out.println(mapEntry.getKey() + " ----- " + mapEntry.getValue());
        }
    }

}
