/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * WordNetUtil.java
 *
 * Created on July 18, 2006, 1:38 PM
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package net.infantai.nlp;

import java.io.FileInputStream;
import java.io.InputStream;
import java.util.GregorianCalendar;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import net.didion.jwnl.JWNL;
import net.didion.jwnl.data.IndexWord;
import net.didion.jwnl.data.IndexWordSet;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.dictionary.MorphologicalProcessor;



/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class WordNetUtil 
{
    
    private static WordNetUtil wordnetUtil = null;
    
    private static net.didion.jwnl.dictionary.Dictionary wordnetDictionary = null;
    
    private static IWLLoader iwlLoader = null;
    private static final Set<String> nonWordNetWords = new LinkedHashSet<>();
    
    private static final Object singletonLock = new Object();
    
    /* Creates a new instance of WordNetUtil */
    private WordNetUtil() 
    {
        try
        {
            InputStream jwnlProperties = new FileInputStream("./jwnl-properties.xml");;
            JWNL.initialize(jwnlProperties);
            wordnetDictionary = net.didion.jwnl.dictionary.Dictionary.getInstance();
            
            iwlLoader = IWLLoader.getInstance();
            nonWordNetWords.addAll(iwlLoader.getPrepositions());
            nonWordNetWords.addAll(iwlLoader.getConjunctions());
            nonWordNetWords.addAll(iwlLoader.getPronouns());
            nonWordNetWords.addAll(iwlLoader.getInterjections());
            nonWordNetWords.addAll(iwlLoader.getArticles());
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
    }
    
    /* 
     * Get an instance (Singleton Object) of class WordNetUtil.
     */
    public static WordNetUtil getInstance()
    {
        synchronized (singletonLock)
        {
            if (wordnetUtil == null)
            {
                wordnetUtil = new WordNetUtil();
            }        
        }
        return wordnetUtil;
    }
    
    
    public String getLemmaWordForm(String strTerm, POS partOfSpeechForm)
    {
        String strLemma = null;
        try
        {
            IndexWord iWord = wordnetDictionary.lookupIndexWord(partOfSpeechForm, strTerm);
            if (iWord != null)
            {
                strLemma = iWord.getLemma();
            }
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return strLemma;
    }    
    
    public String getLemmaWordAnyForm(String strTerm)
    {
        String strLemma = null;
        List listAllPOS = null;
        String inputTermLCase = strTerm.toLowerCase();
        try
        {
            listAllPOS = POS.getAllPOS(); // Gets POS objects labelled NOUN, VERB, ADJECTIVE, ADVERB only.
            IndexWord iWord;
            for (int i = 0; i < listAllPOS.size(); ++i)
            {
                iWord = wordnetDictionary.lookupIndexWord((POS)listAllPOS.get(i), strTerm);
                if (iWord != null)
                {
                    strLemma = iWord.getLemma();
                    break;
                }
            }
            
            if (strLemma == null)
            {
                // Custom word list check for Prepositions, Conjunctions, Pronouns, Interjections, and Articles
                if (nonWordNetWords.contains(inputTermLCase))
                {
                    strLemma = inputTermLCase;
                }            
            }
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return strLemma;
    }
    
    public boolean isInDictionary(String strTerm)
    {
        boolean foundInDictionary = false;
        try
        {
            IndexWordSet iWordSet = wordnetDictionary.lookupAllIndexWords(strTerm);
            if (iWordSet != null)
            {
                if (iWordSet.size() > 0)
                {
                    foundInDictionary = true;
                }
            }
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return foundInDictionary;
    }
    
    public IndexWord getDictionaryWordForm(String strTerm, POS partOfSpeechForm)
    {
        IndexWord iWord = null;
        try
        {
            iWord = wordnetDictionary.lookupIndexWord(partOfSpeechForm, strTerm);
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return iWord;
    }    
    
    public IndexWordSet getAllIndexWords(String strTerm)
    {
        IndexWordSet iWordSet = null;
        try
        {
            iWordSet = wordnetDictionary.lookupAllIndexWords(strTerm);
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return iWordSet;
    }    
    
    public IndexWordSet getVerbTenses(String strTerm)
    {
        IndexWordSet iWordSet = null;
        try
        {
            MorphologicalProcessor mpro = wordnetDictionary.getMorphologicalProcessor();
            List<String> baseForms = mpro.lookupAllBaseForms(POS.VERB, strTerm);
            for (int i = 0; i < baseForms.size(); ++i)
            {
                String s = baseForms.get(i);
                System.out.println(s);
            }
            iWordSet = wordnetDictionary.lookupAllIndexWords(strTerm);
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return iWordSet;
    }    
    
    
    public static void main(String[] args)
    {
        
        String s = "my";
        WordNetUtil wnu = WordNetUtil.getInstance();
        IndexWordSet iws = wnu.getAllIndexWords(s);
        IndexWord[] iwa = iws.getIndexWordArray();
        for (int i = 0; i < iwa.length; ++i)
        {
            IndexWord iw = iwa[i];
            System.out.println("----Index Word: " + iw.toString());
        }
        wnu.getVerbTenses(s);
        System.out.println("preposition: " + wnu.getLemmaWordAnyForm(s));
    }
    
}