/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.security;

import java.util.Base64;
import java.util.GregorianCalendar;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class Console
{
    
    private static final int MINIMUM_KEY_LENGTH_AES = 8; // Characters
    private static final int MINIMUM_KEY_LENGTH_3DES = 24; // Characters
    private static final int MAX_KEY_LENGTH = 128;
    
    public static final String DEFAULT_CHARSET = "ISO-8859-1";
    public static final String UTF8_CHARSET = "UTF-8";

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args)
    {
        try
        {
            
            System.out.println("\n" + new java.util.GregorianCalendar().toZonedDateTime() + " " + "====== Running CryptoUtil ======");
            
            String plaintext = null;
            String ciphertext = null;
            String key = null;
            String algo = null;
            
            boolean testmode = false;
            
            for (int i = 0; i < args.length; ++i)
            {
                if (args[i].equalsIgnoreCase("-plaintext"))
                {
                    plaintext = args[++i];
                }
                else if (args[i].equalsIgnoreCase("-ciphertext"))
                {
                    ciphertext = args[++i];
                }
                else if (args[i].equalsIgnoreCase("-key"))
                {
                    key = args[++i];
                }
                else if (args[i].equalsIgnoreCase("-algo"))
                {
                    algo = args[++i].toUpperCase();
                }                
                else if (args[i].equalsIgnoreCase("-test"))
                {
                    testmode = true;
                }
            }
            
            if (algo == null || key == null)
            {
                showUsage();
                return;
            }
            
            if (plaintext == null && ciphertext == null)
            {
                System.err.println("\nError: Either -plaintext, or -ciphertext must be specified. Use -plaintext to encrypt, -ciphertext to decrypt.");
                showUsage();
                return;
            }          
            
            if (plaintext != null && ciphertext != null)
            {
                System.err.println("\nError: Parameter -plaintext cannot be specified with parameter -ciphertext. Use -plaintext to encrypt, -ciphertext to decrypt.");
                showUsage();
                return;
            }            
            
            if (key.length() > MAX_KEY_LENGTH)
            {
                System.err.println("\nError: Key cannot be more than " + MAX_KEY_LENGTH + " characters in length.");
                return;
            }
            
            byte[] encBytes = null;
            String b64enc = null;
            byte[] b64dec = null;
            byte[] decryptedBytes = null;
            
            switch (algo) // Yes, switch-case can be used with String objects in Java 7 and above. Comparisons will be case-sensitive.
            {
                case "AES":
                    if (key.length() < MINIMUM_KEY_LENGTH_AES)
                    {
                        System.err.println("\nError: AES Key should be at least " + MINIMUM_KEY_LENGTH_AES + " characters in length.");
                        return;
                    }
                    
                    if (ciphertext == null)
                    {
                        // Encryption mode.
                        byte[] inputBytes = plaintext.getBytes(DEFAULT_CHARSET); // plaintext is guaranteed to not be null here
                        encBytes = CryptoUtil.encrypt2(inputBytes, key);
                        b64enc = Base64.getEncoder().encodeToString(encBytes); // ISO-8859-1
                    }
                    else
                    {
                        b64enc = ciphertext; // Decryption mode; assume that input passed on the command-line is Base64 encoded.
                    }
                    b64dec = Base64.getDecoder().decode(b64enc);
                    decryptedBytes = CryptoUtil.decrypt2(b64dec, key);
                    break;
                case "3DES":
                case "TRIPLEDES":                    
                case "DESEDE":
                    if (key.length() < MINIMUM_KEY_LENGTH_3DES)
                    {
                        System.err.println("\nError: 3DES Key should be at least " + MINIMUM_KEY_LENGTH_3DES + " characters in length.");
                        return;
                    }
                    if (ciphertext == null)
                    {
                        byte[] inputBytes = plaintext.getBytes(DEFAULT_CHARSET); // plaintext is guaranteed to not be null here
                        encBytes = CryptoUtil.encrypt1(inputBytes, key);
                        b64enc = Base64.getEncoder().encodeToString(encBytes); // ISO-8859-1                          
                    }
                    else
                    {
                        b64enc = ciphertext; // Decryption mode; assume that input passed on the command-line is Base64 encoded.
                    }    
                    b64dec = Base64.getDecoder().decode(b64enc);
                    decryptedBytes = CryptoUtil.decrypt1(b64dec, key);                    
                    break;
                default:
                    System.err.println("\nError: Unrecognized Cipher Name (" + algo + "): Pass the name of a supported Cipher for the algo command-line parameter.");
                    showUsage();
                    return;
            }
            
            if (b64enc != null && decryptedBytes != null)
            {
                System.out.println(new java.util.GregorianCalendar().toZonedDateTime() + " => " + algo + " Encrypted Bytes (Base64 Encoded):" + b64enc);
                if (testmode || ciphertext != null)
                {
                    System.out.println(new java.util.GregorianCalendar().toZonedDateTime() + " => " + "Decrypted Bytes:" + new String(decryptedBytes, DEFAULT_CHARSET));
                }            
            }
            
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
            showUsage();
        }        
    }
    
    private static void showUsage()
    {
        String usage = "\nCryptoUtil v1.0 (August 2016) - Command-Line Utility to encrypt any arbitrary input string using the key specified. \n";
        usage = usage + "Author: Ashish Nawal Saharia (ashish.saharia@yahoo.co.in) \n";
        usage = usage + "\nUsage: \n";
        usage = usage + "-algo: Cipher to be used for encryption / decryption. (3DES | AES) \n";
        usage = usage + "-plaintext: The input string to be encrypted. \n";
        usage = usage + "-ciphertext: Base64-encoded bytes (encrypted) to be decrypted. \n";
        usage = usage + "-key: The Key to be used for symmetric encryption. \n";
        usage = usage + "-test: Run in test mode, include the original (decrypted) string in the output. \n";
        System.err.println(usage);
    }    
    
}
