/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.security;

import java.io.ByteArrayOutputStream;
import java.security.AlgorithmParameters;
import java.security.SecureRandom;
import java.security.spec.KeySpec;
import java.util.Arrays;
import java.util.GregorianCalendar;
import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESedeKeySpec;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class CryptoUtil
{
    
    private static final String ENCRYPTION_ALGORITHM_3DES = "DESede";     
    private static final String TRANSFORMATION_3DES = "DESede/CBC/PKCS5Padding";
    
    private static final String KEY_DERIVATION_ALGORITHM_AES = "PBKDF2WithHmacSHA1"; // Password Based Key Derivation Function 2 (PBKDF2) / Hash-based Message Authentication code (HMAC)
    private static final String ENCRYPTION_ALGORITHM_AES = "AES";    
    private static final int ITERATION_COUNT_AES = 32768;
    private static final int KEY_LENGTH_AES = 128;
    private static final String TRANSFORMATION_AES = "AES/CBC/PKCS5Padding"; // Algorithm/Feedback Mode/Padding Scheme - AES / Cipher Block Chaining / Public Key Cryptography Standards (PKCS5/PKCS7) padding *Note: What Java calls PKCS5Padding, is actually PKCS7 padding.    

    /**
     * Encrypts input using Triple-DES.
     * @param input - Byte array to be encrypted.
     * @param key - Encryption key (symmetric) to be used.
     * @return Encrypted byte array, or null if the encryption fails due to any exception.
     */
    public static byte[] encrypt1(byte[] input, String key)
    {
        byte[] encryptedBytes = null;
        try
        {
            SecureRandom srandom = new SecureRandom();
            byte[] saltBytes = new byte[16];
            srandom.nextBytes(saltBytes);

            SecretKeyFactory factory = SecretKeyFactory.getInstance(ENCRYPTION_ALGORITHM_3DES);
            DESedeKeySpec desedekeyspec = new DESedeKeySpec(key.getBytes());
            SecretKey secretKey = factory.generateSecret(desedekeyspec);
            SecretKeySpec secretKeySpec = new SecretKeySpec(secretKey.getEncoded(), ENCRYPTION_ALGORITHM_3DES);

            Cipher cipher = Cipher.getInstance(TRANSFORMATION_3DES);
            cipher.init(Cipher.ENCRYPT_MODE, secretKeySpec);
            AlgorithmParameters algoparams = cipher.getParameters();
            byte[] iv = algoparams.getParameterSpec(IvParameterSpec.class).getIV();
            byte[] encryptedData = cipher.doFinal(input);

            // concatenate salt + iv + ciphertext
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            outputStream.write(saltBytes);
            outputStream.write(iv);
            outputStream.write(encryptedData);

            encryptedBytes = outputStream.toByteArray();
            
            outputStream.close();
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return encryptedBytes;
    }        
    
    /**
     * Decrypts the input byte array encrypted with Triple-DES.
     * @param encryptedBytes Byte array encrypted with Triple-DES. Must include 
     * Salt and IV also.
     * @param key Key used for encryption. Triple-DES is symmetric, so the 
     * same key must be used for encryption and decryption.
     * @return Byte array of the original (decrypted) data.
     */
    public static byte[] decrypt1(byte[] encryptedBytes, String key)
    {
        byte[] decryptedBytes = null;
        try
        {
            byte[] salt = Arrays.copyOfRange(encryptedBytes, 0, 16);
            byte[] iv = Arrays.copyOfRange(encryptedBytes, 16, 24);
            byte[] ct = Arrays.copyOfRange(encryptedBytes, 24, encryptedBytes.length);

            SecretKeyFactory factory = SecretKeyFactory.getInstance(ENCRYPTION_ALGORITHM_3DES);
            DESedeKeySpec desedekeyspec = new DESedeKeySpec(key.getBytes());
            SecretKey secretKey = factory.generateSecret(desedekeyspec);
            SecretKeySpec secretKeySpec = new SecretKeySpec(secretKey.getEncoded(), ENCRYPTION_ALGORITHM_3DES);
            Cipher cipher = Cipher.getInstance(TRANSFORMATION_3DES);

            cipher.init(Cipher.DECRYPT_MODE, secretKeySpec, new IvParameterSpec(iv));
            decryptedBytes = cipher.doFinal(ct);
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return decryptedBytes;
    }    
    
    /**
     * Encrypts input using AES.
     * @param input - Byte array to be encrypted.
     * @param key - Encryption key (symmetric) to be used.
     * @return Encrypted byte array, or null if the encryption fails due to any exception.
     */    
    public static byte[] encrypt2(byte[] input, String key)
    {
        byte[] encryptedBytes = null;
        try
        {
            SecureRandom srandom = new SecureRandom();
            byte[] saltBytes = new byte[16];
            srandom.nextBytes(saltBytes);

            SecretKeyFactory factory = SecretKeyFactory.getInstance(KEY_DERIVATION_ALGORITHM_AES);
            KeySpec spec = new PBEKeySpec(key.toCharArray(), saltBytes, ITERATION_COUNT_AES, KEY_LENGTH_AES);
            SecretKey secretKey = factory.generateSecret(spec);
            SecretKeySpec secretKeySpec = new SecretKeySpec(secretKey.getEncoded(), ENCRYPTION_ALGORITHM_AES);

            Cipher cipher = Cipher.getInstance(TRANSFORMATION_AES);
            cipher.init(Cipher.ENCRYPT_MODE, secretKeySpec);
            AlgorithmParameters algoparams = cipher.getParameters();
            byte[] iv = algoparams.getParameterSpec(IvParameterSpec.class).getIV(); // Initialization Vector (IV) length for AES will be 16 bytes (128 bits) - same as the block-size of the cipher, not the key size.
            byte[] encryptedData = cipher.doFinal(input);

            // concatenate salt + iv + ciphertext
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            outputStream.write(saltBytes);
            outputStream.write(iv);
            outputStream.write(encryptedData);

            encryptedBytes = outputStream.toByteArray();
            
            outputStream.close();
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return encryptedBytes;
    }    
    
    /**
     * Decrypts the input byte array encrypted with AES.
     * @param encryptedBytes Byte array encrypted with AES. Must include 
     * Salt and IV also.
     * @param key Key used for encryption. AES is symmetric, so the 
     * same key must be used for encryption and decryption.
     * @return Byte array of the original (decrypted) data.
     */    
    public static byte[] decrypt2(byte[] encryptedBytes, String key)
    {
        byte[] decryptedBytes = null;
        try
        {
            byte[] salt = Arrays.copyOfRange(encryptedBytes, 0, 16);
            byte[] iv = Arrays.copyOfRange(encryptedBytes, 16, 32);
            byte[] ct = Arrays.copyOfRange(encryptedBytes, 32, encryptedBytes.length);

            SecretKeyFactory factory = SecretKeyFactory.getInstance(KEY_DERIVATION_ALGORITHM_AES);
            KeySpec spec = new PBEKeySpec(key.toCharArray(), salt, ITERATION_COUNT_AES, KEY_LENGTH_AES);
            SecretKey secretKey = factory.generateSecret(spec);
            SecretKeySpec secretKeySpec = new SecretKeySpec(secretKey.getEncoded(), ENCRYPTION_ALGORITHM_AES);
            Cipher cipher = Cipher.getInstance(TRANSFORMATION_AES);

            cipher.init(Cipher.DECRYPT_MODE, secretKeySpec, new IvParameterSpec(iv));
            decryptedBytes = cipher.doFinal(ct);
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();            
        }
        return decryptedBytes;
    }    
    
}
