/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.util.common;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.GregorianCalendar;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 *
 * This class contains methods that allow flexible Boolean Text Rules to be 
 * matched against any arbitrary input text. Here, a Boolean Text Rule is a String 
 * containing lookup-tokens with logical AND, OR, and NOT operators in a C-style 
 * syntax. Currently, there is no support for wildcard characters. This class is 
 * intended to match single words or short phrases (as rules), not entire sentences 
 * or paragraphs; any part of a BooleanTextRule that has more than 
 * MAX_TOKENS_IN_GROUP (default = 6) consecutive tokens will never yield a 
 * positive match, even if the input text contains the exact same String. 
 * To avoid such unintended false negatives, the matching function 
 * - matchBooleanTextRule(Set&lt;String&gt; inputTokens) - will throw a 
 * MalformedExpressionException, if it detects (on a best-effort basis) more 
 * than MAX_TOKENS_IN_GROUP consecutive tokens in any part of the Rule.
 * <br>
 * The following operators can be used in a BooleanTextRule:
 * <br>
 * <b>&&</b> - Logical AND
 * <br>
 * <b>||</b> - Logical OR
 * <br>
 * <b>!</b> &nbsp;- Logical Complement (NOT)
 * <br>
 * Tokens and operators can be grouped using parentheses, much like C / Java 
 * boolean constructs. 
 * <br>
 * Automatic Stemming / Lemmatization will <b>not</b> be performed; 
 * "connect" and "connects" will be considered as two different terms.
 * <br><br>
 * <b>Restrictions:</b>
 * <ul>
 * <li>Character ╪ [Extended ASCII Code 216 (Decimal) / UTF-8 0xE2 0x95 0xAA] 
 * cannot be a part of a BooleanTextRule.</li>
 * <li>Character ≡ [Extended ASCII Code 240 (Decimal) / UTF-8 0xE2 0x89 0xA1] 
 * cannot be a part of a BooleanTextRule.</li>
 * <li>When using negation (Logical Complement / NOT), there should be no Space
 * between the NOT operator (!) and the sub-expression being negated.</li>
 * <li>When using negation (Logical Complement / NOT), there should be a Space
 * between the NOT operator (!) and any AND (&&) / OR (||) operator appearing 
 * immediately before it.</li>
 * <li>A rule cannot contain the following Unicode / UTF-8 Space Characters: 
 * 0xC2 0xA0 (No-Break Space), 0xE2 0x80 0xAF (Narrow No-Break Space), 
 * 0xE2 0x80 0x88 (Punctuation Space), and 0xE2 0x81 0x9F (Medium Mathematical Space).</li>
 * <li>A rule cannot contain Newline characters [ASCII Code 13 (Decimal), ASCII Code 10 (Decimal)].</li>
 * <li>A rule cannot contain consecutive Exclamation (!) [ASCII Code 33 (Decimal)] characters.</li>
 * <li>The Vertical Pipe character (|) [ASCII Code 124 (Decimal)] cannot be a part 
 * of any <i>token</i> in the rule.</li> 
 * <li>Currently, there is no provision to escape reserved characters.</li> 
 * <li>This class is primarily designed to match Text Rules against regular English 
 * language input; behaviour could be inconsistent or even broken with 
 * other languages.</li> 
 * </ul>
 * Currently, a MalformedExpressionException is thrown only from 
 * <i>matchBooleanTextRule(Set&lt;String&gt; inputTokens)</i> if the rule cannot 
 * be parsed (the constructor does not throw any Exception). This behaviour 
 * can change in a later release, and the constructor might throw this 
 * exception too.
 * <br><br>
 * <b>Examples:</b> 
 * <br>
 * <blockquote>
 * Set&lt;String&gt; inputTextTokenCombinations = BooleanTextRuleMatcher.getDistinctTokenCombinations("Bring me my Bow of burning gold; Bring me my Arrows of desire: Bring me my Spear...");
 * <br>
 * BooleanTextRuleMatcher btrm1 = new BooleanTextRuleMatcher("bring && my bow && !(longbow || prize)"); 
 * <br>
 * BooleanTextRuleMatcher btrm2 = new BooleanTextRuleMatcher("prize && (draw a longbow || arrow)"); 
 * <br>
 * btrm1.matchBooleanTextRule(inputTextTokenCombinations); // Returns true.
 * <br>
 * btrm2.matchBooleanTextRule(inputTextTokenCombinations); // Returns false.
 * </blockquote>
 * <br>
 * <blockquote>
 * BooleanTextRuleMatcher btrm = new BooleanTextRuleMatcher("prize && (draw a longbow || arrow)");
 * <br>
 * btrm.matchBooleanTextRule("Bring me my Bow of burning gold; Bring me my Arrows of desire: Bring me my Spear...");  // Returns false.
 * <br>
 * btrm.matchBooleanTextRule("And everyone was bidden that could draw a longbow, and the prize was to be an arrow of pure beaten gold.");  // Returns true.
 * </blockquote>
 * 
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class BooleanTextRuleMatcher
{
    
    private static final String TOKEN_DELIM_1 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".“”()[]{}‘’․…⋮‐‑‒–—∼∽〜`~|&'\\/?¿╪¦=!¡+^#*%©®™_$¤¢€£₹¥†‡¶§∙◀▶▲▼◁▷△▽●■◆☑☒⌘�›‹«»〉》《〈@∂∫∬∮∯∑ΣσαβγφΦθϴλδΔζρχΨωΩμπτ≃≅≊≈≠≤≥°Åℇℯ¬∧∨⋀⋁⊻⊼⊽⊨⊭∶∷∩∪⋂⋃±＋﹢−×⋅⨯÷∿∇∆∟⦜⊾∠∡∢⌒⌔⌓∥∁∀∈∉⊂⊃⊄⊅⊆⊈⊇⊉∅⊕⊗⨁⨂⌖∏∐⨉′″√∛∝∞⟵⟶←→↑↓↔↺↻↶↷⇤⇥⇒⇔⇎⟸⟹⟺≪≫≡∴∵🌌☉♁🌊☵☱☲☳☴☶☷☇☈⚡☄☾☽♆♌♒♨♂♀♮♩♪♫♬♯♭♔♕♖♗♘♙♚♛♜♝♞♟☸☯⚙⚛☌☍⎈⚓⛵🚤🚢✈🚁🚆🚄☹☺❄☃☁☂☔★⋆☆﹏☼∃∄"; // Splits on common special characters. Note that 🌌 (Milky Way) (UTF-8 0xF0 0x9F 0x8C 0x8C) and 🌊 (Water Wave) (UTF-8 0xF0 0x9F 0x8C 0x8A) are currently the only 4-byte Unicode / UTF-8 characters in this set. Can be driven through a config file, but this is more fun.
    private static final String TOKEN_DELIM_2 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".“”[]()?‘’․…⋮‐‑‒–—"; // Does not split on "&", "!".
    private static final String TOKEN_DELIM_3 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~|&'\".“”[]/‘’․…⋮‐‑‒–—"; // Does not split on "-", "(", ")".
    private static final String TOKEN_DELIM_4 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~|'\".[]/()?‘’․…⋮‐‑‒–—"; // Does not split on "-", but splits on "(", ")".
    private static final String TOKEN_DELIM_5 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~|'\"[]/-!()?‘’…⋮‐‑‒–—"; // Does not split on ".", useful for tokens like "amazon.com".
    private static final String TOKEN_DELIM_6 = " <>	    \r\n";
    private static final String TOKEN_DELIM_7 = " <>	    \r\n“”‘’.0123456789";
    private static final List<Set<Character>> ALL_DELIM_CHARACTERS_SETS = new ArrayList<Set<Character>>();
    
    private static final String TEXT_RULE_DELIM = "&|"; // WARNING: Do not split on space here !
    private static final String DOUBLE_AMPERSAND_LOGICAL_AND = "&&";
    private static final String DOUBLE_AMPERSAND_LOGICAL_AND_REPLACEMENT = "╪╪";
    private static final char AMPERSAND = '&';
    private static final char RESERVED_AMPERSAND_REPLACEMENT = '≡';
    private static final char RESERVED_LOGICAL_AND_REPLACEMENT = '╪';
    private static final char EXCLAMATION_LOGICAL_NOT = '!';
    private static final char CARRIAGE_RETURN = (char) 13;
    private static final char LINE_FEED = (char) 10;
    private static final String RULE_BOOLEAN_PART_TRUE = "true"; // Warning: Do not change this to UPPER CASE, since there is no call to toLowerCase() in BooleanExpressionParser.parseExpression(). 
    private static final String RULE_BOOLEAN_PART_FALSE = "false"; // Warning: Do not change this to UPPER CASE, since there is no call to toLowerCase() in BooleanExpressionParser.parseExpression().    
    
    private static final Tokenizer TOKENIZER = new Tokenizer();

    /**
     * Mutable static... I know, I know. 
     * I had kept this value a constant (final field, set to 6) originally, but 
     * now think that programmers (users of this library) should be able to modify 
     * it if they choose to do so. Should probably move the static tokenization 
     * methods to a new, separate class, but don't want to break 
     * backwards-compatibility yet. MAX_TOKENS_IN_GROUP still defaults to 6, 
     * and can be set only before this class is instantiated or used for 
     * tokenization.
     * <br>
     * See Also:
     * {@link net.infantai.na.util.common.BooleanTextRuleMatcher#setMaxTokensToCombineForAllRules(int maxTokensToCombine) setMaxTokensToCombineForAllRules(int maxTokensToCombine)}
     */
    private static int MAX_TOKENS_IN_GROUP = 6;    
    private static volatile boolean freezeValueOfMaxTokensInGroup = false;
    
    private static final char[] CHARACTERS_TO_TRIM = new char[] {'(', ')', ' '};
    private static final char[] CHARACTERS_TO_TRIM_LEFT = new char[] {'!'};
    
    private String textRule = null;
    private String originalTextRule = "";
    private List<TokenGroup> allTextRuleTokens = new ArrayList<TokenGroup>();
    
    private boolean isSimpleKeyword = false;
    
    private boolean isMalformedExpression = false;
    private String malformedExpressionExceptionMessage = "";
    
    private boolean caseSensitive = false;
    
    static
    {
        
        if (ALL_DELIM_CHARACTERS_SETS.isEmpty())
        {
            
            // In Java, static initialization is performed only once per class-loader, 
            // so synchronization isn't really needed here. In environments with 
            // multiple class-loaders, this class (like any other class), with 
            // all its static members etc., will be maintained separately and independently 
            // by each class-loader that loads it.

            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_1));
            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_2));
            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_3));
            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_4));
            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_5));
            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_6));
            ALL_DELIM_CHARACTERS_SETS.add(CommonUtil.stringToSetOfCharacters(TOKEN_DELIM_7));

        }
        
    }
    
    /**
     * Constructor to create regular (<b>case-insensitive</b>) Boolean Text Rules. 
     * Since <i>case-sensitive</i> rules are rarely needed, the use of this constructor 
     * should be preferred over <i>BooleanTextRuleMatcher(String booleanTextRule, boolean caseSensitive)</i>.
     * @param booleanTextRule 
     */
    public BooleanTextRuleMatcher(String booleanTextRule)
    {
        initialize(booleanTextRule.toLowerCase());   
    }
    
    /**
     * Constructor to create case-sensitive Boolean Text Rules. Use this 
     * constructor if this rule needs to perform only case-sensitive matches. 
     * @param booleanTextRule
     * @param caseSensitive 
     */
    public BooleanTextRuleMatcher(String booleanTextRule, boolean caseSensitive)
    {
        if (!caseSensitive)
        {
            initialize(booleanTextRule.toLowerCase());
        }
        else
        {
            initialize(booleanTextRule);
            this.caseSensitive = true;
        }
    }
    
    private void initialize(String booleanTextRule)
    {
        freezeValueOfMaxTokensInGroup = true;
        
        originalTextRule = booleanTextRule;
        textRule = originalTextRule;
        textRule = textRule.replace(DOUBLE_AMPERSAND_LOGICAL_AND, DOUBLE_AMPERSAND_LOGICAL_AND_REPLACEMENT); //        textRule = textRule.replace("&&", "╪╪");
        textRule = textRule.replace(AMPERSAND, RESERVED_AMPERSAND_REPLACEMENT); //        textRule = textRule.replace("&", "≡");
        textRule = textRule.replace(DOUBLE_AMPERSAND_LOGICAL_AND_REPLACEMENT, DOUBLE_AMPERSAND_LOGICAL_AND); //        textRule = textRule.replace("╪╪", "&&");               
        allTextRuleTokens = TOKENIZER.getAllTokens(textRule, TEXT_RULE_DELIM, 1);
        if (allTextRuleTokens.size() == 1)
        {
            String tgToken = allTextRuleTokens.get(0).getToken();
            if (tgToken.indexOf('!') == -1 && tgToken.indexOf("&&") == -1 && tgToken.indexOf("||") == -1)
            {
                isSimpleKeyword = true;
            }
        }
        
        if (!isMalformedExpression)
        {
            checkRule(originalTextRule);
        }    
    }
    
    private void checkRule(String originalTextRule)
    {
        boolean exclamationSpace = false;
        boolean foundValidOperator = false;
        boolean expectedAndOr = false;
        char lastChar = (char) 0;
        char lastNonSpaceChar = (char) 0;
        int lastNonSpaceCharIndex = -1;
        int lastValidOperatorIndex = -1;
        int consecutiveTokenCount = 0;
        boolean incrementConsecutiveTokenCount = false;
        boolean foundValidTextTokenPart = false;
        Deque<Integer> leftBracketIndexStack = new ArrayDeque<Integer>();
        StringBuilder sbLeftSubstring = new StringBuilder();
        for (int i = 0; i < originalTextRule.length(); ++i)
        {
            char ch = originalTextRule.charAt(i);
            switch (ch)
            {
                case ' ': // Unicode spaces can easily be processed as delimiters. But just for fun (and some sanity), let's not allow them in rules.
                    isMalformedExpression = true;
                    malformedExpressionExceptionMessage = "Rule cannot contain UTF-8 Character 0xC2 0xA0 (No-Break Space) (index " + i + ").";
                    return;
//                    break;
                case ' ':
                    isMalformedExpression = true;
                    malformedExpressionExceptionMessage = "Rule cannot contain UTF-8 Character 0xE2 0x80 0xAF (Narrow No-Break Space) (index " + i + ").";
                    return;
//                    break;
                case ' ':
                    isMalformedExpression = true;
                    malformedExpressionExceptionMessage = "Rule cannot contain UTF-8 Character 0xE2 0x80 0x88 (Punctuation Space) (index " + i + ").";
                    return;
//                    break;                    
                case ' ':
                    isMalformedExpression = true;
                    malformedExpressionExceptionMessage = "Rule cannot contain UTF-8 Character 0xE2 0x81 0x9F (Medium Mathematical Space) (index " + i + ").";
                    return;
//                    break;                                        
                case ' ':
                    if (lastNonSpaceChar == EXCLAMATION_LOGICAL_NOT)
                    {
                        exclamationSpace = true;
                    }
                    else
                    {
                        exclamationSpace = false;                    
                    }
                    incrementConsecutiveTokenCount = true;
                    break;
                case EXCLAMATION_LOGICAL_NOT:
//                    if (lastChar == EXCLAMATION_LOGICAL_NOT)
//                    {
//                        isMalformedExpression = true;
//                        malformedExpressionExceptionMessage = "Consecutive Exclamations !! (Index " + (i - 1) + ")";                        
//                        return;
//                    }
//                    if (foundValidOperator)
//                    {
//                        if (i < (originalTextRule.length() - 1))
//                        {
//                            char nextChar = originalTextRule.charAt(i + 1);
//                            if (nextChar == ' ')
//                            {
//                                isMalformedExpression = true;
//                                malformedExpressionExceptionMessage = "Space not allowed between logical NOT (!) and sub-expression after index " + i + ".";
//                                return;                                
//                            }
//                        }                        
//                    }
                    break;
                case '&':
                    if (lastChar == '&')
                    {
                        
                        if (i < (originalTextRule.length() - 1))
                        {
                            char nextChar = originalTextRule.charAt(i + 1);
                            if (nextChar == '&' || nextChar == '|' || nextChar == '!') // Do not allow "&&&", "&&|", and "&&!".
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "Incorrect operator at index " + (i - 1) + ".";
                                return;                                
                            }
                            else if (sbLeftSubstring.toString().trim().equals("&"))
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "No operand before operator at index " + (i - 1) + ".";
                                return;
                            }                            
                            else
                            {
                                foundValidOperator = true;
                                expectedAndOr = false;
                                
                                consecutiveTokenCount = 0;
                                incrementConsecutiveTokenCount = false;
                                
                                foundValidTextTokenPart = false;
                                lastValidOperatorIndex = i - 1;
                            }
                        }
                        else
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "No operand after operator at index " + (i - 1) + ".";
                            return;
                        }
                    }
                    break;
                case '|':
                    if (lastChar == '|')
                    {
                        if (i < (originalTextRule.length() - 1))
                        {
                            char nextChar = originalTextRule.charAt(i + 1);
                            if (nextChar == '|' || nextChar == '&' || nextChar == '!') // Do not allow "|||", "||&", and "||!".
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "Incorrect operator at index " + (i - 1) + ".";
                                return;                                
                            }
                            else if (sbLeftSubstring.toString().trim().equals("|"))
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "No operand before operator at index " + (i - 1) + ".";
                                return;
                            }                            
                            else
                            {
                                foundValidOperator = true;
                                expectedAndOr = false;
                                
                                consecutiveTokenCount = 0;
                                incrementConsecutiveTokenCount = false;
                                
                                foundValidTextTokenPart = false;
                                lastValidOperatorIndex = i - 1;
                            }
                        }
                        else
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "No operand after operator at index " + (i - 1) + ".";
                            return;                                                            
                        }
                    }
                    else if (lastChar == '&')
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Incorrect operator at index " + (i - 1) + ".";
                        return;                                                                                
                    }
                    else
                    {
                        if (i < (originalTextRule.length() - 1))
                        {
                            char nextChar = originalTextRule.charAt(i + 1);
                            if (nextChar != '|')
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "Incorrect operator at index " + i + ".";
                                return;                                                            
                            }
                        }
                        else // end of rule
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "Incorrect operator at index " + i + ".";
                            return;
                        }
                    }
                    break;
                case '(':
                    if (exclamationSpace) 
                    {
                        if (foundValidOperator) // Allow "Yahoo! (India)" but not "Yahoo && ! (India)".
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "Space not allowed between logical NOT operator (!) and sub-expression near index " + i + ".";
                            return;
                        }
                        else if (sbLeftSubstring.toString().trim().equals("!")) // Begins with logical NOT (!) followed by a space...
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "Space not allowed immediately after logical NOT operator (!) (before sub-expression near index " + i + ").";
                            return;
                        }
                    }
                    if (i > 0 && lastChar != '(' && lastChar != '!' && !foundValidOperator && lastNonSpaceChar == ')') // For cases like "(abc || def) (xyz)".
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Missing or misplaced operator before left bracket at index " + i + ".";
                        return;
                    }
                    else if (lastChar == '!' && !foundValidOperator && !sbLeftSubstring.toString().trim().equals("!")) // For cases like "(abc || def) !(xyz)".
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Missing or misplaced operator before sub-expression at index " + (i - 1) + ".";
                        return;                        
                    }
                    else if (expectedAndOr)
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Operator expected before left bracket at index " + i + ".";                                                            
                        return;                        
                    }                    
                    
                    foundValidOperator = false;
                    leftBracketIndexStack.push(new Integer(i));
                    foundValidTextTokenPart = false;
                    break;
                case RESERVED_AMPERSAND_REPLACEMENT:
                    isMalformedExpression = true;
                    malformedExpressionExceptionMessage = "Reserved Character: Rule cannot contain " + RESERVED_AMPERSAND_REPLACEMENT + " (index " + i + ").";                                            
                    return;
//                    break;
                case RESERVED_LOGICAL_AND_REPLACEMENT:
                    if (lastChar == RESERVED_LOGICAL_AND_REPLACEMENT)
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Reserved Character Sequence: Rule cannot contain " + DOUBLE_AMPERSAND_LOGICAL_AND_REPLACEMENT + " (index " + (i - 1) + ").";                                                            
                        return;
                    }
                    break;
                case ')':
                    if (leftBracketIndexStack.isEmpty())
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Mismatched ')' at index " + i + ".";                                                            
                        return;
                    }
                    if (lastNonSpaceChar == '|')
                    {
                        if (lastNonSpaceCharIndex > 0)
                        {
                            char chBeforeVerticalPipe = originalTextRule.charAt(lastNonSpaceCharIndex - 1);
                            if (chBeforeVerticalPipe == '|')
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "No operand after operator at index " + (lastNonSpaceCharIndex - 1) + ".";                                                            
                                return; 
                            }
                        }                        
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Unacceptable operator at index " + lastNonSpaceCharIndex + ".";                                                            
                        return;                    
                    }
                    if (lastNonSpaceChar == '&')
                    {
                        if (lastNonSpaceCharIndex > 0)
                        {
                            char chBeforeAmpersand = originalTextRule.charAt(lastNonSpaceCharIndex - 1);
                            if (chBeforeAmpersand == '&')
                            {
                                isMalformedExpression = true;
                                malformedExpressionExceptionMessage = "No operand after operator at index " + (lastNonSpaceCharIndex - 1) + ".";                                                            
                                return; 
                            }
                        }
                    }                    
                    
                    foundValidOperator = false;
                    Integer lastLeftBracketIndex = leftBracketIndexStack.pop();
                    
                    boolean foundOperatorBeforeLastLeftBracket = false;
                    boolean breakLoop = false;
                    if (lastLeftBracketIndex > 0)
                    {
                        for (int j = (lastLeftBracketIndex - 1); j >= 0; --j)
                        {
                            char c = originalTextRule.charAt(j);
                            switch (c)
                            {
                                case '&':
                                    if (originalTextRule.charAt((j - 1)) == '&')
                                    {
                                        foundOperatorBeforeLastLeftBracket = true;
                                        breakLoop = true;
                                        break;
                                    }                                    
                                    break;
                                case '|':
                                    if (originalTextRule.charAt((j - 1)) == '|')
                                    {
                                        foundOperatorBeforeLastLeftBracket = true;
                                        breakLoop = true;
                                        break;
                                    }
                                    break;
                                case '!':
                                case ' ':
                                    break;
                                default:
                                    foundOperatorBeforeLastLeftBracket = false;
                                    breakLoop = true;
                                    break;
                            }
                            if (breakLoop)
                            {
                                break;
                            }
                        }
                        if (foundOperatorBeforeLastLeftBracket)
                        {
                            expectedAndOr = true;
                        }                    
                    }
                    
                    if (!foundValidTextTokenPart)
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "No valid text token found in sub-expression before right bracket at index " + i + ".";                                                            
                        return;                                                
                    }
                    break;
                case CARRIAGE_RETURN:
                case LINE_FEED:
                    isMalformedExpression = true;
                    malformedExpressionExceptionMessage = "Rule cannot contain Newline characters [ASCII Code 13 (Decimal), ASCII Code 10 (Decimal)].";                                                            
                    return;                                            
//                    break;
                default:
                    foundValidOperator = false;
                    foundValidTextTokenPart = true;
                    if (expectedAndOr)
                    {
                        isMalformedExpression = true;
                        malformedExpressionExceptionMessage = "Operator expected before sub-expression at index " + i + ".";                                                            
                        return;                        
                    }
                    else
                    {
                        if (exclamationSpace && sbLeftSubstring.toString().trim().equals("!"))
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "Space not allowed between logical NOT (!) and sub-expression near index " + i + ".";
                            return;
                        }                    
                    }
                    
                    if (incrementConsecutiveTokenCount && lastChar == ' ')
                    {
                        if (consecutiveTokenCount >= MAX_TOKENS_IN_GROUP)
                        {
                            isMalformedExpression = true;
                            malformedExpressionExceptionMessage = "Rule part cannot contain more than " + MAX_TOKENS_IN_GROUP + " consecutive tokens. Check rule near index " + i + ".";
                            return;                            
                        }                        
                        consecutiveTokenCount++;
                        incrementConsecutiveTokenCount = false;
                    }
                    
                    break;
            }
            if (ch != ' ')
            {
                lastNonSpaceChar = ch;
                lastNonSpaceCharIndex = i;
            }
            lastChar = ch;
            sbLeftSubstring.append(ch);
        }
        
        if (!leftBracketIndexStack.isEmpty())
        {
            Integer mismatchedLeftBracketIndex = leftBracketIndexStack.pop();
            isMalformedExpression = true;
            malformedExpressionExceptionMessage = "Mismatched '(' at index " + mismatchedLeftBracketIndex + ".";
        }
        else if (!foundValidTextTokenPart)
        {
            if (lastValidOperatorIndex > -1)
            {
                isMalformedExpression = true;
                malformedExpressionExceptionMessage = "No valid text token found after operator at index " + lastValidOperatorIndex + ".";                    
            }
            else
            {
                isMalformedExpression = true;
                malformedExpressionExceptionMessage = "Rule does not contain any valid text token.";                    
            }
        }
        
    }
    
    /**
     * Convenience method. Use 
     * matchBooleanTextRule(Set&lt;String&gt; articleTokens) 
     * instead, by invoking getDistinctTokenCombinations(String input) separately first, 
     * if the same input needs to be matched against multiple BooleanTextRuleMatcher 
     * objects. Absorbs any Exception thrown by the underlying functions.
     * 
     * @param input The input text against which this rule will be matched.
     * @return A boolean value - true if this Boolean Text Rule matches the 
     * input, false if it does not.
     */
    public boolean matchBooleanTextRule(String input)
    {
        boolean matched = false;
        try
        {
            Set<String> articleTokens = null;
            if (!this.caseSensitive)
            {
                articleTokens = getDistinctTokenCombinations(input);
            }
            else
            {
                articleTokens = getDistinctTokenCombinationsPreserveCase(input);
            }
            matched = matchBooleanTextRule(articleTokens);        
        }
        catch (MalformedExpressionException meex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + meex.getMessage());
            meex.printStackTrace();
            System.err.println("Rule: " + this.toString());
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return matched;
    }
    
    /**
     * Returns true if this Boolean Text Rule satisfies the input tokens, else 
     * returns false. The value returned by <b>getDistinctTokenCombinations(String input)</b> 
     * should be used as input to this function. When the same input String needs to 
     * be matched against multiple BooleanTextRule objects, make a single call 
     * to <b>getDistinctTokenCombinations(String input)</b> to first get all 
     * distinct tokens, then call the matchBooleanTextRule(Set&lt;String&gt; articleTokens) 
     * method of different BooleantextRuleMatcher objects. 
     * @param inputTokens - A Set of Strings as obtained by a call to method 
     * getDistinctTokenCombinations(String input) of this class. 
     * @return A boolean value - true if this Boolean Text Rule matches the 
     * input, false if it does not.
     * @throws net.infantai.util.common.MalformedExpressionException 
     */
    public boolean matchBooleanTextRule(Set<String> inputTokens) throws MalformedExpressionException
    {
        boolean matched = false;
        
        if (isMalformedExpression)
        {
            throw new MalformedExpressionException(malformedExpressionExceptionMessage);
        }
         
        String originalRule = textRule.replace(RESERVED_AMPERSAND_REPLACEMENT, AMPERSAND); //            String originalSearchString = textRule.replace("≡", "&");
        List<TokenGroup> ruleTGList = allTextRuleTokens;

        // Return quickly if this rule is a simple keyword, and not a real 
        // boolean rule with logical constructs.
        if (isSimpleKeyword)
        {
            String simpleKeyword = allTextRuleTokens.get(0).getToken().replace(RESERVED_AMPERSAND_REPLACEMENT, AMPERSAND);
            simpleKeyword = trimBracketsBalanced(simpleKeyword);
            return inputTokens.contains(simpleKeyword);
        }
        //

        List<String> ruleBooleanParts = new ArrayList<String>();

        for (int i = 0; i < ruleTGList.size(); ++i)
        {
            TokenGroup tgRule = ruleTGList.get(i);

            String ruleToken = tgRule.getToken().trim();
            String ruleTokenWhitespaceTrimmed = ruleToken;

            if (ruleToken.length() == 0)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Error in Rule: " + this.toString());
                String pre = "";
                String post = "";
                if (i > 0)
                {
                    pre = ruleTGList.get(i - 1).getToken().trim();
                }
                if (i < (ruleTGList.size() - 1))
                {
                    post = ruleTGList.get(i + 1).getToken().trim();
                }
                throw new MalformedExpressionException("Incorrect input near \"" + pre + " ... " + post + "\". Check operator.");
            }

            boolean checkCharacterSpaceBracket = false;
            String ruleTokenProcessed = ruleToken;
            do 
            {
                ruleToken = ruleTokenProcessed;
                if (containsMatchingBrackets(ruleTokenProcessed))
                {
                    int rulePartLengthPre = ruleTokenProcessed.length();
                    ruleTokenProcessed = trimBracketsBalanced(ruleTokenProcessed);
                    int rulePartLengthPost = ruleTokenProcessed.length();
                    if (rulePartLengthPre == rulePartLengthPost)
                    {
                        checkCharacterSpaceBracket = false;
                    }
                }
                else
                {
                    ruleTokenProcessed = CommonUtil.trimCharacters(CHARACTERS_TO_TRIM, ruleTokenProcessed, 1, 1);
                    checkCharacterSpaceBracket = true;
                }
                ruleTokenProcessed = CommonUtil.trimCharactersLeft(CHARACTERS_TO_TRIM_LEFT, ruleTokenProcessed);
            }
            while (ruleTokenProcessed.length() < ruleToken.length());            
            
            if (checkCharacterSpaceBracket)
            {
                if (foundCharacterSpaceBracket(ruleTokenProcessed))
                {
                    String errMsg = "";
                    int missingOperatorIndex = originalTextRule.indexOf(ruleTokenWhitespaceTrimmed);
                    if (missingOperatorIndex != -1)
                    {
                        missingOperatorIndex = originalTextRule.indexOf("(", missingOperatorIndex);
                        if (missingOperatorIndex != -1)
                        {
                            errMsg = "Missing operator before left bracket at index " + missingOperatorIndex + ".";
                        }
                    }
                    if (errMsg.length() == 0)
                    {
                        String post = null;
                        if (i < (ruleTGList.size() - 1))
                        {
                            post = ruleTGList.get(i + 1).getToken().trim();
                        }
                        errMsg = "Incorrect input near \" ... " + ruleTokenWhitespaceTrimmed;
                        if (post != null)
                        {
                            errMsg = errMsg + " ... " + post;
                        }
                        errMsg = errMsg + "\". Check rule token.";                    
                    }
                    throw new MalformedExpressionException(errMsg);                    
                }
            }

            if (ruleToken.length() == 0)
            {
                String post = null;
                if (i < (ruleTGList.size() - 1))
                {
                    post = ruleTGList.get(i + 1).getToken().trim();
                }
                String errMsg = "Incorrect input near \" ... " + ruleTokenWhitespaceTrimmed;
                if (post != null)
                {
                    errMsg = errMsg + " ... " + post;
                }
                errMsg = errMsg + "\". Check rule token.";
                throw new MalformedExpressionException(errMsg);
            }
            else
            {
                char ruleTokenWhitespaceTrimmedFirstChar = ruleTokenWhitespaceTrimmed.charAt(0);
                if (ruleTokenWhitespaceTrimmedFirstChar == '(' || ruleTokenWhitespaceTrimmedFirstChar == '!' || ruleTokenWhitespaceTrimmedFirstChar == ')') // startsWith
                {
                    int from = 0;
                    int to = 0;
                    from = tgRule.getIndex();
                    if (ruleTokenWhitespaceTrimmed.equals("(")) // Can happen if the rule is malformed.
                    {
                        to = 1;
                    }
                    else
                    {
                        to = tgRule.getIndex() + tgRule.getToken().indexOf(ruleToken);
                    }
                    ruleBooleanParts.add(originalRule.substring(from, to));
                }            
            }


            if (inputTokens.contains(ruleToken.replace(RESERVED_AMPERSAND_REPLACEMENT, AMPERSAND)))
            {
                ruleBooleanParts.add(RULE_BOOLEAN_PART_TRUE);
            }
            else
            {
                ruleBooleanParts.add(RULE_BOOLEAN_PART_FALSE);
            }

            if (ruleTokenWhitespaceTrimmed.charAt((ruleTokenWhitespaceTrimmed.length() - 1)) == ')') // endsWith
            {
                int from = tgRule.getIndex() + tgRule.getToken().indexOf(ruleToken) + ruleToken.length();
                int to = tgRule.getIndex() + tgRule.getSize();
                ruleBooleanParts.add(originalRule.substring(from, to));
            }

            if (i < (ruleTGList.size() - 1))
            {
                String oper = originalRule.substring(tgRule.getIndex() + tgRule.getToken().length(), ruleTGList.get(i + 1).getIndex());
                ruleBooleanParts.add(oper);
            }
        }

        StringBuilder sbBooleanExp = new StringBuilder();
        for (int i = 0; i < ruleBooleanParts.size(); ++i)
        {
            if (i != 0)
            {
                sbBooleanExp.append(' ');
            }
            sbBooleanExp.append(ruleBooleanParts.get(i));
        }

        try
        {
//            System.out.println("Expr: " + sbBooleanExp.toString());
            matched = BooleanExpressionParser.parseExpression(sbBooleanExp.toString());
        }
        catch (MalformedExpressionException meex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Error parsing Rule: " + this.toString());
            throw meex;
        }
        return matched;
    }
    
    private boolean foundCharacterSpaceBracket(String ruleToken)
    {
        boolean foundCharacter = false;
        boolean foundCharacterSpace = false;
        boolean foundCharacterSpaceBracket = false;
//                    for (int j = 0; j < ruleTokenProcessed.length(); ++j)
        for (int j = 0; j < ruleToken.length(); ++j)
        {
            char ch = ruleToken.charAt(j);
            switch (ch)
            {
                case ' ':
                    if (foundCharacter)
                    {
                        foundCharacterSpace = true;
                    }
                    break;
                case '(':
                    if (foundCharacterSpace)
                    {
                        foundCharacterSpaceBracket = true;
                    }
                    break;
                default:
                    foundCharacter = true;
                    foundCharacterSpace = false;
                    break;
            }
            if (foundCharacterSpaceBracket)
            {
                break;
            }
        }
        if (foundCharacterSpaceBracket)
        {
            return true;
//            throw new MalformedExpressionException("Incorrect input near \"" + ruleToken.replace(RESERVED_AMPERSAND_REPLACEMENT, AMPERSAND) + "\". Missing operator before left bracket.");
        }
        return false;
    }    
    
    /**
     * Method to trim '(' (left bracket) and ')' (right bracket) from both, the 
     * left and the right of the input string, only if they are matched. In other 
     * words, if the right-most character of the input String is a right 
     * bracket, it will be trimmed (discarded) only if the left-most character 
     * of the input String is a left bracket.
     * <br>
     * <b>Examples:</b>
     * <br>
     * trimBracketsBalanced("(The Galaxy)"); // returns "The Galaxy"
     * <br>
     * trimBracketsBalanced("Mumbai (India)"); // returns "Mumbai (India)"
     * @param ruleToken
     * @return 
     */
    private String trimBracketsBalanced(String ruleToken)
    {
        String trimmedString = ruleToken.trim();
        
        if (trimmedString == null || trimmedString.length() == 0)
        {
            return trimmedString;
        }
        
        char chLeft = trimmedString.charAt(0);
        while (chLeft == '(')
        {
            char chRight = trimmedString.charAt(trimmedString.length() - 1);
            if (chRight == ')')
            {
                trimmedString = trimmedString.substring(1, trimmedString.length() - 1).trim();
                if (trimmedString.length() == 0)
                {
                    return trimmedString;
                }
                chLeft = trimmedString.charAt(0);            
            }
            else
            {
                break;
            }
        }

        return trimmedString.trim();
    }    
    
    /**
     * This method is used to determine if unmatched brackets in rule tokens 
     * should be trimmed or not. We typically want to trim the left bracket from 
     * the rule token "(thanks", but not the right bracket from the rule token 
     * "Mumbai (India)". The value returned by this method is used to determine if 
     * method trimBracketsBalanced(String ruleToken) should be used on the input 
     * ruleToken or not. Basically, if this method returns false, brackets 
     * should be trimmed from the input ruleToken even if they are unmatched.
     * Returns true if all left brackets are matched with 
     * corresponding right brackets. Returns false if no brackets are found. 
     * Returns false if brackets are mismatched. 
     * <br>
     * See Also: 
     * {@link net.infantai.na.util.common.BooleanTextRuleMatcher#trimBracketsBalanced(String ruleToken) trimBracketsBalanced(String ruleToken)}
     * <br><br>
     * Examples: 
     * <br>
     * containsMatchingBrackets("Mumbai (India)"); // returns true
     * <br>
     * containsMatchingBrackets("(The Galaxy)"); // returns true
     * <br>
     * containsMatchingBrackets("(Guide to the"); // returns false
     * <br>
     * containsMatchingBrackets("thanks for all the fish)"); // returns false
     * @param ruleToken
     * @return 
     */
    private boolean containsMatchingBrackets(String ruleToken)
    {
        boolean matchingBracketsFound = false;
        Deque<Integer> leftBracketIndexStack = new ArrayDeque<Integer>();
        int leftBracketCount = 0;
        for (int i = 0; i < ruleToken.length(); ++i)
        {
            char ch = ruleToken.charAt(i);
            if (ch == '(')
            {
                leftBracketIndexStack.push(new Integer(i));
                leftBracketCount++;
            }
            else if (ch == ')')
            {
                if (!leftBracketIndexStack.isEmpty())
                {
                    leftBracketIndexStack.pop();
                }
                else
                {
                    return false;
                }
            }
        }
        if (leftBracketIndexStack.isEmpty() && leftBracketCount > 0)
        {
            matchingBracketsFound = true;
        }
        return matchingBracketsFound;
    }    
    
    /**
     * When the same input String needs to be matched against multiple 
     * BooleanTextRule objects, first use this method once to get all input 
     * tokens, then call the matchBooleanTextRule(Set&lt;String&gt; articleTokens) 
     * method of different BooleantextRuleMatcher objects. 
     * Using <i>matchBooleanTextRule(String input)</i> will internally invoke this 
     * method every single time, thereby incurring an unnecessary computational 
     * cost. This method converts the input String to lower case before tokenizing it.
     * 
     * @param input The input text against which Boolean Text Rules will be matched.
     * @return A Set&lt;String&gt; containing all distinct tokens (and token combinations) 
     * of the input string. 
     */    
    public static Set<String> getDistinctTokenCombinations(String input)
    {
        
        freezeValueOfMaxTokensInGroup = true;
        
        Set<String> distinctTokens = new LinkedHashSet<String>();
        String inputLCase = input.trim().toLowerCase();
        
        for (int j = 0; j < ALL_DELIM_CHARACTERS_SETS.size(); ++j)
        {
            distinctTokens.addAll(TOKENIZER.getDistinctTokens(inputLCase, ALL_DELIM_CHARACTERS_SETS.get(j), MAX_TOKENS_IN_GROUP));
        }
        
        return distinctTokens;
    }
    
    /**
     * When the same input String needs to be matched against multiple 
     * <b>case-sensitive</b> rules, first use this method once to get all input 
     * tokens, then call the matchBooleanTextRule(Set&lt;String&gt; articleTokens) 
     * method of different BooleantextRuleMatcher objects. 
     * Using <i>matchBooleanTextRule(String input)</i> will internally invoke this 
     * method every single time, thereby incurring an unnecessary computational 
     * cost. This method preserves the case of the input String, and should be 
     * used <b>only</b> for tokenizing input text for case-sensitive rules. 
     * <br>
     * If there are no case-sensitive rules in the system, this method is not 
     * needed at all. If the system has a mix of both, case-sensitive, 
     * and case-insensitive rules, the input text should be tokenized twice - 
     * once with <i>getDistinctTokenCombinations(String input)</i>, and once with 
     * <i>getDistinctTokenCombinationsPreserveCase(String input)</i>. 
     * The output of <i>getDistinctTokenCombinations(String input)</i> should be 
     * passed to the <i>matchBooleanTextRule(Set&lt;String&gt; articleTokens)</i> 
     * method of case-insensitive (default) rules, and the output of 
     * <i>getDistinctTokenCombinationsPreserveCase(String input)</i> should be 
     * passed to the <i>matchBooleanTextRule(Set&lt;String&gt; articleTokens)</i> 
     * method of case-sensitive rules.
     * <br>
     * <b>WARNING</b>: Do not use this method to tokenize input text for 
     * regular (case-insensitive) rules. 
     * 
     * @param input The input text against which Boolean Text Rules will be matched.
     * @return A Set&lt;String&gt; containing all distinct tokens (and token combinations) 
     * of the input string. 
     */    
    public static Set<String> getDistinctTokenCombinationsPreserveCase(String input)
    {
        
        freezeValueOfMaxTokensInGroup = true;
        
        Set<String> distinctTokens = new LinkedHashSet<String>();
        
        for (int j = 0; j < ALL_DELIM_CHARACTERS_SETS.size(); ++j)
        {
            distinctTokens.addAll(TOKENIZER.getDistinctTokens(input, ALL_DELIM_CHARACTERS_SETS.get(j), MAX_TOKENS_IN_GROUP));
        }
        
        return distinctTokens;
    }

    /**
     * Returns MAX_TOKENS_IN_GROUP. This value limits the maximum number of tokens 
     * that will be combined as a single String, when method 
     * <i>getDistinctTokenCombinations(String input)</i> is called on some input text.
     * @return MAX_TOKENS_IN_GROUP
     */
    public static int getMaxTokensInGroupForAllRules()
    {
        return MAX_TOKENS_IN_GROUP;
    }
    
    /**
     * Sets MAX_TOKENS_IN_GROUP. This value limits the maximum number of tokens 
     * that will be combined as a single String, when method 
     * <i>getDistinctTokenCombinations(String input)</i> is called on some input text. 
     * Consequently, it also limits the maximum number of consecutive tokens to 
     * be used in any part of a BooleanTextRule. The default value is 6. If the 
     * default value needs to be changed, this method must be invoked <b>before</b> 
     * any call has been made to the constructor of this class 
     * (BooleanTextRuleMatcher), or to method getDistinctTokenCombinations(String input). 
     * Calling this method after this class has been instantiated even once, 
     * or after a call to method getDistinctTokenCombinations(String input) has been made, 
     * will result in a java.lang.UnsupportedOperationException being thrown.
     * <br><br>
     * <b>WARNING</b>: Setting this value will change the MAX_TOKENS_IN_GROUP 
     * value for <b>all</b> rules.
     * <br>
     * See Also:
     * {@link net.infantai.na.util.common.BooleanTextRuleMatcher#getDistinctTokenCombinations(String) getDistinctTokenCombinations(String input)}
     * @param maxTokensToCombine - Maximum number of tokens to combine as a 
     * single String when method getDistinctTokenCombinations(String input) of 
     * this class is called.
     * @throws java.lang.UnsupportedOperationException 
     */
    public static void setMaxTokensToCombineForAllRules(int maxTokensToCombine)
    {
        if (freezeValueOfMaxTokensInGroup)
        {
            throw new UnsupportedOperationException("Value of MAX_TOKENS_IN_GROUP cannot be changed if BooleanTextRuleMatcher has been instantiated even once, or has been used for tokenization.");
        }
        MAX_TOKENS_IN_GROUP = maxTokensToCombine;
    }
    

    /**
     * Method to determine if this BooleanTextRule is case-sensitive or not.
     * @return True if this was instantiated as a case-sensitive rule, false otherwise. 
     */
    public boolean isCaseSensitive()
    {
        return caseSensitive;
    }    
    
    @Override
    public String toString()
    {
        return originalTextRule;
    }
    
}
